class HeightWidth:
    def __init__(self, height: int, width: int):
        self.height = height
        self.width = width


class Padding:
    def __init__(self, top: int, bottom: int, left: int, right: int):
        self.top = top
        self.bottom = bottom
        self.left = left
        self.right = right


class HorizontalPadding:
    def __init__(self, first: int, second: int):
        self.first = first
        self.second = second
